import { EmojiResolvable, MessageButtonStyle, Message, MessageSelectOptionData, MessageEmbed, MessageEmbedOptions, User, TextBasedChannels, Guild, GuildMember, GuildChannel, Role, Client } from 'discord.js';

declare type ConsoleLoggerLevel = "info" | "success" | "error" | "warn";

/**
 * Options for {@link buttonComponent}
 */
declare type ButtonComponentOptions = {
    /**
     * The text to display on the button
     */
    label: string;
    /**
     * An emoji displayed on the button.
     * If you only want to show an emoji, pass an empty string for the label.
     * @see https://discord.js.org/#/docs/main/stable/typedef/EmojiResolvable
     */
    emoji?: EmojiResolvable;
    /**
     * The color and intent of the button.
     * @see https://discord.js.org/#/docs/main/stable/typedef/MessageButtonStyle
     */
    style: Exclude<MessageButtonStyle, "LINK">;
    /**
     * Whether the button is disabled.
     * This can be anti-accessibility, so consider an alternative,
     * like showing an error on click, or hiding the button entirely.
     */
    disabled?: boolean;
    /**
     * Called when the button is clicked
     */
    onClick: (context: ButtonInteractionContext) => unknown;
};
/**
 * Returned from {@link buttonComponent}
 */
declare type ButtonComponent = ButtonComponentOptions & {
    type: "button";
    customId: string;
};
/**
 * The context object received by button onClick handlers.
 * See {@link buttonComponent}
 */
declare type ButtonInteractionContext = InteractionContext & {
    readonly message: Message;
};
/**
 * Represents a discord [button](https://discord.com/developers/docs/interactions/message-components#buttons) component.
 * Does not support URL or disabled props yet.
 *
 * ```js
 * context.reply(() => (
 *   buttonComponent({
 *     label: "Click me!",
 *     onClick: context => {
 *       context.reply(() => "You clicked me!"),
 *     },
 *   }),
 * ))
 * ```
 */
declare function buttonComponent(options: ButtonComponentOptions): ButtonComponent;

/**
 * Options for the link component
 * @see linkComponent
 */
declare type LinkComponentOptions = {
    /**
     * The text to display on the button
     */
    label: string;
    /**
     * An emoji displayed on the button.
     * If you only want to show an emoji, pass an empty string for the label.
     * @see https://discord.js.org/#/docs/main/stable/typedef/EmojiResolvable
     */
    emoji?: EmojiResolvable;
    /**
     * The URL to open when the button is clicked
     */
    url: string;
};
/**
 * Returned from {@link linkComponent}
 */
declare type LinkComponent = LinkComponentOptions & {
    type: "link";
};
/**
 * Creates a link component, which is a button that opens a URL when clicked.
 */
declare function linkComponent(options: LinkComponentOptions): LinkComponent;

/**
 * Options passed to {@link selectMenuComponent}
 */
declare type SelectMenuComponentOptions = {
    /**
     * The array of options that can be selected.
     * Same structure as [MessageSelectOptionData from DJS](https://discord.js.org/#/docs/main/stable/typedef/MessageSelectOptionData)
     */
    options: MessageSelectOptionData[];
    /**
     * The currently selected option value(s).
     * This accepts `Iterable`, so you can pass an array, a Set,
     * or any other kind of iterable.
     * @see selectMenuComponent
     */
    selected?: Iterable<string> | string | undefined;
    /** The placeholder text to display when no options are selected */
    placeholder?: string | undefined;
    /**
     * Called when one or more options are selected.
     * Use this callback to update your current selected state.
     *
     * **Note:** For multi-select ({@link SelectMenuComponentOptions.maxValues}), this doesn't get called immediately.
     * It only gets called after clicking away from the select dropdown.
     */
    onSelect: (context: SelectMenuInteractionContext) => unknown;
    /**
     * The minimum number of options that can be selected.
     * Passing this option will enable multi-select,
     * and can't be 0.
     */
    minValues?: number | undefined;
    /**
     * The maximum number of options that can be selected.
     * Passing this option will enable multi-select,
     * and can't be greater than the number of options.
     */
    maxValues?: number | undefined;
};
/**
 * Returned from {@link selectMenuComponent}
 */
declare type SelectMenuComponent = Omit<SelectMenuComponentOptions, "selected"> & {
    type: "selectMenu";
    customId: string;
};
/**
 * Context object passed to {@link SelectMenuComponentOptions.onSelect}
 */
declare type SelectMenuInteractionContext = InteractionContext & {
    /** The message that the component is on */
    readonly message: Message;
    /** The values the user selected */
    readonly values: string[];
};
/**
 * Represents a Discord [select menu](https://discord.com/developers/docs/interactions/message-components#select-menus) component.
 *
 * ```js
 * let selected
 *
 * context.reply(() =>
 *   selectMenuComponent({
 *     options: [
 *       { label: "option 1", value: "option 1" },
 *       { label: "option 2", value: "option 2" },
 *     ],
 *     placeholder: "select one",
 *     selected,
 *     onChange: (values) => {
 *       selected = values[0]
 *     },
 *   }),
 * )
 * ```
 */
declare function selectMenuComponent({ options, selected, ...args }: SelectMenuComponentOptions): SelectMenuComponent;

/**
 * Returned from {@link actionRowComponent}
 */
declare type ActionRowComponent = {
    type: "actionRow";
    children: ActionRowChild[];
};
/**
 * A valid child of {@link actionRowComponent}
 */
declare type ActionRowChild = SelectMenuComponent | ButtonComponent | LinkComponent;
/**
 * A component that represents a Discord [action row](https://discord.com/developers/docs/interactions/message-components#action-rows)
 * and follows the same limitations (max 5 buttons, max 1 select, can't mix both).
 *
 * You usually don't have to use this yourself;
 * Gatekeeper will automatically create action rows for you.
 * But if you have a specific structure in mind, you can still use this.
 *
 * ```js
 * context.reply(() => [
 *   // normally, these two buttons would be on the same line,
 *   // but you can use action row components to put them on different lines
 *   actionRowComponent(
 *     buttonComponent({
 *       // ...
 *     }),
 *   ),
 *   actionRowComponent(
 *     buttonComponent({
 *       // ...
 *     }),
 *   ),
 * ])
 * ```
 */
declare function actionRowComponent(...children: Array<ActionRowChild | ActionRowChild[]>): ActionRowComponent;

/**
 * Returned from {@link embedComponent}
 */
declare type EmbedComponent = {
    type: "embed";
    embed: MessageEmbed | MessageEmbedOptions;
};
/**
 * Creates an embed component.
 * Accepts {@link https://discord.js.org/#/docs/main/stable/typedef/MessageEmbedOptions MessageEmbedOptions}, or a DJS {@link https://discord.js.org/#/docs/main/stable/class/MessageEmbed MessageEmbed} instance.
 *
 * ```js
 * context.reply(() => [
 *   embedComponent({
 *     title: "Your weather today 🌤",
 *     description: `Sunny, with a 12% chance of rain`,
 *     footer: {
 *       text: "Sourced from https://openweathermap.org/",
 *     },
 *   }),
 * ])
 * ```
 */
declare function embedComponent(embed: MessageEmbedOptions | MessageEmbed): EmbedComponent;

/**
 * Any reply component object
 */
declare type ReplyComponent = TextComponent | EmbedComponent | ActionRowComponent | ButtonComponent | SelectMenuComponent | LinkComponent;
/**
 * A gatekeeper-specific type representing top-level components,
 * stuff that doesn't need an extra wrapper.
 *
 * For example, a button isn't top level,
 * as it needs to be wrapped in an action row
 */
declare type TopLevelComponent = TextComponent | EmbedComponent | ActionRowComponent;
/**
 * Represents the text in a message
 */
declare type TextComponent = {
    type: "text";
    text: string;
};
/**
 * The function passed to `context.reply`
 */
declare type RenderReplyFn = () => RenderResult;
/**
 * Anything that can be rendered in a reply.
 *   - Embed components, action row components, and text components are accepted as-is
 *   - Button and select menu components are automatically placed in action rows, respecting discord's restrictions
 *   - Strings and numbers become text components. Empty strings can be used to add empty lines in the message. `\n` will also add a new line.
 *   - Nested arrays are flattened
 *   - Everything else (booleans, `undefined`, `null`) is ignored
 */
declare type RenderResult = ReplyComponent | string | number | boolean | undefined | null | RenderResult[];

/** Returned from {@link InteractionContext.reply}, for arbitrarily doing stuff with a reply message */
declare type ReplyHandle = {
    /** The discord message associated with this reply */
    get message(): Message | undefined;
    /**
     * Refresh the reply message.
     *
     * Gatekeeper will call this automatically on a component interaction,
     * but you can do so yourself when you want to update the message
     * outside of a component interaction
     */
    readonly refresh: () => void;
    /** Delete the message. After this point,
     * {@link ReplyHandle.message message} will be undefined */
    readonly delete: () => void;
};
/** Base type for all context objects */
declare type InteractionContext = {
    /** The user that triggered this interaction.
     * For buttons, this is the user that clicked the button,
     * and so on */
    readonly user: User;
    /** The channel that this interaction took place in. */
    readonly channel: TextBasedChannels | undefined;
    /** The guild that this interaction took place in. */
    readonly guild: Guild | undefined;
    /** The guild member for the user that triggered this interaction */
    readonly member: GuildMember | undefined;
    /** Create a new reply for this interaction. This reply can be updated over time via interactions, or via manual {@link ReplyHandle.refresh refresh} calls */
    readonly reply: (render: RenderReplyFn) => ReplyHandle;
    /** Like {@link InteractionContext.reply reply}, but only shows the message to the interacting user.
     *
     * This does not return a reply handle; ephemeral replies can't be updated or deleted manually
     */
    readonly ephemeralReply: (render: RenderReplyFn) => void;
    /** [Defer](https://discordjs.guide/interactions/replying-to-slash-commands.html#deferred-responses) a reply, which shows a loading message. Useful if your command might take a long time to reply. */
    readonly defer: () => void;
    /** Same as {@link InteractionContext.defer defer}, but shows the loading message only to the interacting user. */
    readonly ephemeralDefer: () => void;
};

/**
 * Options for creating a message command. Shows when right-clicking a message.
 * @see Gatekeeper.addMessageCommand
 */
declare type MessageCommandConfig = {
    /**
     * The name of the command. This shows up in the context menu for messages.
     */
    name: string;
    /** Aliases: alternate names to call this command with */
    aliases?: string[];
    /**
     * The function to call when the command is ran.
     */
    run: (context: MessageCommandInteractionContext) => void | Promise<unknown>;
};
/** The context object received when running a message command */
declare type MessageCommandInteractionContext = InteractionContext & {
    /** The message that the command was run on */
    targetMessage: Message;
};

/**
 * Configuration for a slash command. See {@link Gatekeeper.addSlashCommand}
 */
declare type SlashCommandConfig<Options extends SlashCommandOptionConfigMap = SlashCommandOptionConfigMap> = {
    /**
     * The name of the command.
     * e.g. If you pass the name "airhorn",
     * the command will be run with /airhorn in Discord
     */
    name: string;
    /** Aliases: alternate names to call this command with */
    aliases?: string[];
    /**
     * The description of the command.
     * Shows up when showing a list of the bot's commands
     */
    description: string;
    /**
     * An object of options for the command, also called arguments, or parameters.
     */
    options?: Options;
    /**
     * The function to run when the command is called.
     * Receives a {@link SlashCommandInteractionContext} object as the first argument.
     */
    run: (context: SlashCommandInteractionContext<Options>) => void | Promise<unknown>;
};
/**
 * Valid slash command option config, only used for typescript inference
 */
declare type SlashCommandOptionConfigMap = {
    [name: string]: SlashCommandOptionConfig;
};
/**
 * A possible option for a slash command.
 * See {@link SlashCommandOptionValueMap} for a list of possible options
 * and the values they resolve to.
 */
declare type SlashCommandOptionConfig = SlashCommandOptionConfigBase & ({
    type: "STRING";
    choices?: Array<SlashCommandOptionChoiceConfig<string>>;
} | {
    type: "NUMBER" | "INTEGER";
    choices?: Array<SlashCommandOptionChoiceConfig<number>>;
} | {
    type: "BOOLEAN";
} | {
    type: "USER";
} | {
    type: "CHANNEL";
    channelTypes?: SlashCommandOptionChannelType[];
} | {
    type: "ROLE";
} | {
    type: "MENTIONABLE";
});
/**
 * All possible channel types to filter by when using the `CHANNEL` option type
 */
declare type SlashCommandOptionChannelType = "GUILD_TEXT" | "DM" | "GUILD_VOICE" | "GROUP_DM" | "GUILD_CATEGORY" | "GUILD_NEWS" | "GUILD_STORE" | "GUILD_NEWS_THREAD" | "GUILD_PUBLIC_THREAD" | "GUILD_PRIVATE_THREAD" | "GUILD_STAGE_VOICE";
/**
 * A potential choice for a slash command option
 */
declare type SlashCommandOptionChoiceConfig<Value> = {
    name: string;
    value: Value;
};
/**
 * This is the magic that takes your option config
 * and gives you a typesafe object of values.
 */
declare type SlashCommandOptionValues<Options extends SlashCommandOptionConfigMap> = {
    [Name in keyof Options]: Options[Name]["required"] extends true ? SlashCommandOptionValueMap[Options[Name]["type"]] : SlashCommandOptionValueMap[Options[Name]["type"]] | undefined;
};
/**
 * A map of option types to the kind of value it resolves to.
 * e.g. If an option has a type of "NUMBER", it will resolve to a number.
 */
declare type SlashCommandOptionValueMap = {
    STRING: string;
    NUMBER: number;
    INTEGER: number;
    BOOLEAN: boolean;
    USER: User;
    CHANNEL: GuildChannel;
    ROLE: Role;
    MENTIONABLE: SlashCommandMentionableValue;
};
/**
 * A resolved mentionable option for a slash command
 */
declare type SlashCommandMentionableValue = {
    /**
     * A string that can be sent in a message as a mention.
     * e.g. `"<@!123456789>"` for users, `"<@&123456789>"` for roles
     */
    mention: string;
} & ({
    /**
     * Whether or not this mention is a user mention.
     * If using typescript, this property _must_ be checked
     * to use the related properties
     */
    isUser: true;
    /**
     * The mentioned user
     */
    user: User;
    /**
     * The guild (server) member object, if in a guild
     */
    guildMember: GuildMember | undefined;
} | {
    /**
     * Whether or not this mention is a user mention.
     * If using typescript, this property _must_ be checked
     * to use the related properties
     */
    isUser: false;
    /**
     * The role that was mentioned, only available in guilds (servers)
     */
    role: Role;
});
/**
 * The interaction context for a slash command
 */
declare type SlashCommandInteractionContext<Options extends SlashCommandOptionConfigMap = SlashCommandOptionConfigMap> = InteractionContext & {
    /**
     * An object of the options that were passed when running the slash command
     */
    options: SlashCommandOptionValues<Options>;
};
/**
 * Shared properties for all slash command option types
 */
declare type SlashCommandOptionConfigBase = {
    /**
     * Description for the option, shows up when tabbing through the options in Discord
     */
    description: string;
    /**
     * Whether the option is required.
     * If true, the value will be guaranteed to exist in the options object,
     * otherwise it will be undefined
     */
    required?: boolean;
};

/**
 * Options for creating a user command. Shows when right-clicking on a user.
 * @see Gatekeeper.addUserCommand
 */
declare type UserCommandConfig = {
    /**
     * The name of the command. This shows up in the context menu for users.
     */
    name: string;
    /** Aliases: alternate names to call this command with */
    aliases?: string[];
    /**
     * The function to call when the command is ran.
     */
    run: (context: UserCommandInteractionContext) => void | Promise<unknown>;
};
declare type UserCommandInteractionContext = InteractionContext & {
    /** The user that the command was run on */
    readonly targetUser: User;
    /** If in a guild (server), the guild member for the user */
    readonly targetGuildMember: GuildMember | undefined;
};

/** Options for creating a gatekeeper instance */
declare type GatekeeperConfig = {
    /** A Discord.JS client */
    client: Client;
    /** The name of the bot, used for logger messages */
    name?: string;
    /** Show colorful debug logs in the console */
    logging?: boolean | ConsoleLoggerLevel[];
    /**
     * An *absolute path* to a folder with command files.
     * Each file should `export default` a function to accept a gatekeeper instance
     * and add commands to it.
     *
     * ```ts
     * // main.ts
     * Gatekeeper.create({
     *   commandFolder: join(__dirname, "commands"),
     * })
     * ```
     * ```ts
     * // commands/ping.ts
     * export default function addCommands(gatekeeper) {
     *   gatekeeper.addCommand({
     *     name: "ping",
     *     description: "Pong!",
     *     run: (ctx) => ctx.reply(() => "Pong!"),
     *   })
     * }
     * ```
     */
    commandFolder?: string;
    /**
     * Where commands should be registered.
     *
     * `guild` - Register commands in all guilds (servers) that the bot joins, which appear immediately.
     * This is the default, and recommended for testing, or if your bot is just in a few guilds.
     *
     * `global` - Register commands globally, which can take a few minutes or an hour to show up.
     * Since bots have a global command limit, this is good for when your bot grows beyond just a few servers.
     *
     * `both` - Register commands in both guilds and globally.
     *
     * See the discord docs for more info: https://discord.com/developers/docs/interactions/application-commands#registering-a-command
     */
    scope?: "guild" | "global" | "both";
    /**
     * Called when any error occurs.
     * You can use this to log your own errors, send them to an error reporting service, etc.
     */
    onError?: (error: Error) => void;
};
/** Basic information about the commands currently added */
declare type CommandInfo = {
    /** The name of the command */
    name: string;
};
/**
 * A gatekeeper instance.
 * Holds commands, manages discord interactions, etc.
 */
declare class Gatekeeper {
    private readonly config;
    private readonly logger;
    private readonly commands;
    private readonly commandInstances;
    private constructor();
    /** Create a {@link Gatekeeper} instance */
    static create(config: GatekeeperConfig): Promise<Gatekeeper>;
    /** Returns a list of basic info for each added command */
    getCommands(): readonly CommandInfo[];
    /**
     * Add a slash command
     * ```ts
     * gatekeeper.addSlashCommand({
     *   name: "ping",
     *   description: "Pong!",
     *   run: (ctx) => ctx.reply(() => "Pong!"),
     * })
     * ```
     */
    addSlashCommand<Options extends SlashCommandOptionConfigMap>(config: SlashCommandConfig<Options>): void;
    /**
     * Add a user command
     * ```ts
     * gatekeeper.addUserCommand({
     *   name: 'get user color',
     *   run: (ctx) => ctx.reply(() => ctx.targetGuildMember?.color ?? "not in a guild!"),
     * })
     * ```
     */
    addUserCommand(config: UserCommandConfig): void;
    /**
     * Add a message command
     * ```ts
     * gatekeeper.addMessageCommand({
     *   name: 'reverse',
     *   run: (ctx) => {
     *     ctx.reply(() => ctx.targetMessage.content.split("").reverse().join(""))
     *   }
     * })
     * ```
     */
    addMessageCommand(config: MessageCommandConfig): void;
    private addCommands;
    private addEventListeners;
    private loadCommandsFromFolder;
    private syncGuildCommands;
    private syncGlobalCommands;
    private syncCommands;
    private removeUnusedRemoteCommand;
    private updateChangedCommands;
    private removeAllCommands;
    private handleCommandInteraction;
    private handleComponentInteraction;
    private withErrorHandler;
}

export { ActionRowChild, ActionRowComponent, ButtonComponent, ButtonComponentOptions, ButtonInteractionContext, CommandInfo, ConsoleLoggerLevel, EmbedComponent, Gatekeeper, GatekeeperConfig, InteractionContext, LinkComponent, LinkComponentOptions, MessageCommandConfig, MessageCommandInteractionContext, RenderReplyFn, RenderResult, ReplyComponent, ReplyHandle, SelectMenuComponent, SelectMenuComponentOptions, SelectMenuInteractionContext, SlashCommandConfig, SlashCommandInteractionContext, SlashCommandMentionableValue, SlashCommandOptionChoiceConfig, SlashCommandOptionConfig, SlashCommandOptionConfigBase, SlashCommandOptionConfigMap, SlashCommandOptionValueMap, SlashCommandOptionValues, TextComponent, TopLevelComponent, UserCommandConfig, UserCommandInteractionContext, actionRowComponent, buttonComponent, embedComponent, linkComponent, selectMenuComponent };
